//+------------------------------------------------------------------+
//|                                                           AF.mq5 |
//|                                    Copyright (c) 2023, Marketeer |
//|                          https://www.mql5.com/ru/articles/12187/ |
//+------------------------------------------------------------------+
#include <Graphics/Graphic.mqh>
#include <EnumToArray.mqh>

const string AF_ = "AF_";

//+------------------------------------------------------------------+
//| Calculation of activation functions and their derivatives        |
//+------------------------------------------------------------------+
void Calc(const ENUM_ACTIVATION_FUNCTION af, double &x[], double &y[],
   const bool derive = false)
{
   // [-4, +4]/(1 / 10) -> 81
   const int range = 4;
   const int num = range * 10;
   const int size = range * 2 * num + 1;
   const double step = 1.0 / num;
   
   ArrayResize(x, size);
   for(int i = 0; i < size; ++i)
   {
      x[i] = -range + i * step;
   }
   matrix in, out;
   in.Init(1, size);
   out.Init(1, size);
   in.Swap(x);
   if(derive)
   {
      in.Derivative(out, af);
   }
   else
   {
      in.Activation(out, af);
   }
   out.Swap(y);
   in.Swap(x);
}

//+------------------------------------------------------------------+
//| Setup graphic object                                             |
//+------------------------------------------------------------------+
void SetupGraphic(CGraphic *graphic, const string title)
{
   graphic.GapSize(2);
   graphic.HistoryNameWidth(0);
   graphic.HistoryNameSize(0);
   graphic.HistorySymbolSize(0);
   graphic.YAxis().ValuesWidth(20);
   graphic.XAxis().DefaultStep(5);
   graphic.YAxis().DefaultStep(5);
   graphic.CurvePlotAll();
   graphic.FontSet("Consolas", -80, FW_HEAVY);
   graphic.TextAdd(graphic.Width() / 2, 10, title, clrBlack, TA_CENTER | TA_TOP);
   graphic.Update();
}

//+------------------------------------------------------------------+
//| Setup curve object                                               |
//+------------------------------------------------------------------+
void SetupCurve(CCurve *c)
{
   c.LinesWidth(2);
   c.LinesSmooth(true);
   c.LinesSmoothStep(0.05);
   c.LinesSmoothTension(5.0);
}

//+------------------------------------------------------------------+
//| Script program start function                                    |
//+------------------------------------------------------------------+
void OnStart()
{
   CGraphic *graphic[];
   int values[];
   const int n = EnumToArray<ENUM_ACTIVATION_FUNCTION>(values, 0, 100);
   const int width = (int)ChartGetInteger(0, CHART_WIDTH_IN_PIXELS);
   const int height = (int)ChartGetInteger(0, CHART_HEIGHT_IN_PIXELS);
   
   int nx = (int)MathPow(n, 1.0 / (1.0 + 1.0 * height / width));
   int ny = (int)MathPow(n, 1.0 / (1.0 + 1.0 * width / height));
   int delta = n - nx * ny;
   if(delta > 0)
   {
      if(nx > ny) ny += delta / nx + !!(delta % nx);
      else nx += delta / ny + !!(delta % ny);
   }
   const int dx = width / nx;
   const int dy = height / ny;
   
   ArrayResize(graphic, n);
   for(int i = 0; i < n; ++i)
   {
      const int px = i % nx;
      const int py = i / nx;
      string title = EnumToString((ENUM_ACTIVATION_FUNCTION)i);
      graphic[i] = new CGraphic();
      graphic[i].Create(0, title, 0,
         px * dx, py * dy, (px + 1) * dx, (py + 1) * dy);
      double x[], y[];
      Calc((ENUM_ACTIVATION_FUNCTION)i, x, y);
      SetupCurve(graphic[i].CurveAdd(x, y, CURVE_LINES, ""));
      Calc((ENUM_ACTIVATION_FUNCTION)i, x, y, true);
      SetupCurve(graphic[i].CurveAdd(x, y, CURVE_LINES, ""));
      StringReplace(title, AF_, "   ");
      StringToLower(title);
      SetupGraphic(graphic[i], title);
   }

   while(!IsStopped())
   {
      Sleep(1000);
   }
   
   for(int i = 0; i < n; ++i)
   {
      delete graphic[i];
   }
   
   // NB: CGraphic object does NOT delete underlying on-screen object
   ObjectsDeleteAll(0, AF_);
}
//+------------------------------------------------------------------+
